/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File server.h */

/* Thus speaketh ye mighty CVS:
 * $Id: server.h,v 1.5 1999/05/29 15:56:05 david Exp $ */

/* This file contains the declarations for server.c */

#ifndef _LEGENDES_SERVER_H
#define _LEGENDES_SERVER_H

/* Include files */

#include <stdio.h>		/* Standard I/O */

#include <errno.h>		/* Error numbers */
#include <string.h>		/* String functions such as strerror() */

#include <sys/socket.h>		/* Socket types */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <pthread.h>		/* Posix threads */

/* Debug level */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0
#endif

/* Socket(s) and all */

#ifndef PORTNUM
#define PORTNUM 8888		/* The TCP port we listen on */
#endif

/* The ``global'' socket is the socket which we create with socket()
 * and where we listen() for new connections.  Each new accept()ed
 * connection gives birth to a new, ``local'' socket.  See the
 * accept(2) man page for details. */
extern int gsock;		/* The global socket descriptor */

#ifndef BUFSIZE
#define BUFSIZE 1024		/* The connection buffer size */
#endif

/* lsock_tab is the table of connections and lsock_nb is the number of
 * connections (i.e. the length of lsock_tab).  Each entry in the
 * lsock_tab is a *pointer* to a structure (a struct lsock_s) which
 * describes the connection.  Each lsock_s structure contains a local
 * socket descriptor, a data buffer, a few other associated variables,
 * and very importantly a mutex protecting all these data (see
 * NOTE0001 below for an explanation of how these mutexes work). */
extern int lsock_nb;		/* Number of connections */
extern struct lsock_s {
  pthread_t listener;		/* Listener thread for that connection */
  pthread_t handler;		/* Handler thread for that connection */
  pthread_mutex_t d_mutex;	/* Mutex protecting all the data below */
  char sopen;			/* Whether s is open */
  int s;			/* Local socket descriptor */
  struct sockaddr_in remote;	/* Socket on remote host */
  char handwork;		/* Handler thread has some work to do */
  pthread_cond_t h_cond;	/* Condition used to wake the handler thead */
  enum {
    CONN_STATE_NEW,
    CONN_STATE_NORMAL,
    CONN_STATE_DEAD
  } state;			/* Connection state */
  int buflen;			/* In bytes */
  char buffer[BUFSIZE];		/* Data buffer */
  /* XXX - The following data should probably go in a different
   * structure that gets associated to a character and not to a
   * connection. */
  unsigned long x, y;		/* x and y components of position */
} **lsock_tab; /* Table of local sockets; note ** and not just *... */

extern pthread_mutex_t lsock_mutex;

/* [NOTE0001] The lsock_mutex is the most important mutex in this
 * program.  It protects the lsock_tab.  In other words, a thread is
 * not allowed to modify any entry of the lsock_tab nor its size
 * (i.e. lsock_nb) unless it is holding the lsock_mutex.  For this
 * reason also, many loops on the connections are done with the
 * lsock_mutex held (so the lsock_tab won't be modified under the
 * looper's feet).  This means that the lsock_mutex is a pretty
 * precious resource (and perhaps we should consider replacing it by a
 * rwlock - some profiling obviously needs to be done there).
 * Besides, some other data are protected by the lsock_mutex, such as
 * the connection count, below.  However, it isn't *that* precious in
 * the sense that once a thread has found the pointer (typically
 * called sck in the code) to a connection structure, it need no
 * longer hold the lsock_mutex in order to access the data in that
 * structure.  Indeed, these data are protected by another mutex,
 * namely the d_mutex of the structure.  This d_mutex must be held
 * whenever any data in the structure is to be modified (this includes
 * reading or writing the local socket).  IMPORTANT NOTICE: How does
 * one actually get access to data within a connection?  First, one
 * locks the lsock_mutex and reads the lsock_tab to find the sck
 * pointer to the connection structure.  Then one has to lock the
 * d_mutex.  Now there is a catch there.  (1) The d_mutex must be
 * locked *before* the lsock_mutex is released, because otherwise the
 * connection structure might be free()d and destroyed between the
 * time the lsock_mutex is released and the time the d_mutex is held
 * (for the moment, we don't free() connection structures but that
 * will come, and the only guarantee we make is that the free()ing
 * thread will hold both the lsock_mutex and the d_mutex - so a thread
 * wanting to keep a connection structure must always be holding to
 * one of these mutexes).  But (2) it is FORBIDDEN to
 * pthread_mutex_lock() the d_mutex while the lsock_mutex is held
 * because that may block and the lsock_mutex is too precious to be
 * held by a blocking thread.  So the thread must
 * pthread_mutex_trylock() instead. */

/* Connection count.  The following variable is incremented every time
 * a connection is received or closed or free()d (the latter can't
 * happen yet).  To change conn_count you must first acquire the
 * lsock_mutex. */
extern long conn_count;

/* Useful functions (in conndata.c) */

/* ``Safe'' pthread functions.  I write those because I can't affort
 * to write the error checking clause *every* damn time I lock or
 * unlock a mutex. */

inline extern void
safe_pthread_mutex_init(pthread_mutex_t *mutex,
			const pthread_mutexattr_t *mutexattr)
{
  int error;
  if ((error=pthread_mutex_init(mutex,mutexattr)))
    {
      fprintf(stderr,"Initializing mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline extern void
safe_pthread_mutex_lock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_lock(mutex)))
    {
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline extern int
safe_pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_trylock(mutex)))
    {
      if (error==EBUSY) return EBUSY;
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
  return 0;
}

inline extern void
safe_pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_unlock(mutex)))
    {
      fprintf(stderr,"Unlocking mutex: %s\n",strerror(error));
      exit(1);
    }
}

/* Main functions */

/* In listener.c */
extern void init_gsock (void);
extern void accepter_loop (void);
extern void listener_loop (struct lsock_s *sck);

/* In handler.c */
extern void handler_loop (struct lsock_s *sck);

#endif /* not defined (_LEGENDES_SERVER_H) */
